from threading import local
from unittest.mock import patch, sentinel

from django.template import Engine, Origin, TemplateDoesNotExist, TemplateSyntaxError
from django.template.backends.django import DjangoTemplates
from django.test import TestCase
from testfixtures import TempDirectory

from django_secretary.models import Template as TemplateModel
from django_secretary.utils.templates import Template, TemplateLoader, get_active_template, override_template

from .utils import RenderTestMixin, create_template


class TestGetActiveTemplate(TestCase):
    def setUp(self):
        patcher = patch('django_secretary.utils.templates._ACTIVE_TEMPLATES', local())
        self.addCleanup(patcher.stop)
        self._active_templates = patcher.start()

    def test_no_active(self):
        # Test no active template
        TemplateModel.objects.create(name='example.html', is_active=True)
        with self.assertRaisesMessage(TemplateDoesNotExist, "There is no active template."):
            get_active_template()

    def test_no_active_name(self):
        # Test no active template with the name.
        template = TemplateModel.objects.create(name='example.html', is_active=True)
        self.assertEqual(get_active_template('example.html'), template)

    def test_no_active_not_found(self):
        # Test no active template
        with self.assertRaisesMessage(TemplateDoesNotExist, "'unknown.html' is neither active, nor found in database."):
            get_active_template('unknown.html')

    def test_active(self):
        self._active_templates.template = sentinel.template
        self.assertEqual(get_active_template(), sentinel.template)

    def test_active_name(self):
        self._active_templates.templates = {'example.html': sentinel.template}
        self.assertEqual(get_active_template('example.html'), sentinel.template)

    def test_active_unknown(self):
        self._active_templates.templates = {'example.html': sentinel.template}
        with self.assertRaisesMessage(TemplateDoesNotExist, "'unknown.html' has no active template."):
            get_active_template('unknown.html')


class TestOverrideTemplate(TestCase):
    def setUp(self):
        patcher = patch('django_secretary.utils.templates._ACTIVE_TEMPLATES', local())
        self.addCleanup(patcher.stop)
        self._active_templates = patcher.start()

    def test_override(self):
        # Test override correctly changes template there and back
        template = TemplateModel(name='example.html')
        with override_template(template):
            overriden = get_active_template()
            overriden_name = get_active_template('example.html')

        self.assertEqual(overriden, template)
        self.assertEqual(overriden_name, template)
        # Check template is deactivated when override exited.
        self.assertRaises(TemplateDoesNotExist, get_active_template)
        self.assertRaises(TemplateDoesNotExist, get_active_template, 'example.html')

    def test_override_raise(self):
        # Test override correctly changes template back even if error occurs
        template = TemplateModel(name='example.html')
        with self.assertRaisesMessage(Exception, "ERROR_IN_OVERRIDE"):
            with override_template(template):
                raise Exception("ERROR_IN_OVERRIDE")

        # Check template is deactivated when override exited.
        self.assertRaises(TemplateDoesNotExist, get_active_template)  # type: ignore[unreachable]
        self.assertRaises(TemplateDoesNotExist, get_active_template, 'example.html')

    def test_override_nested(self):
        # Test override correctly changes template there and back, even when nested
        template = TemplateModel(name='example.html')
        nested_template = TemplateModel(name='example.html')
        with override_template(template):
            overriden = get_active_template()
            overriden_name = get_active_template('example.html')
            with override_template(nested_template):
                overriden_nested = get_active_template()
                overriden_nested_name = get_active_template('example.html')
            returned = get_active_template()
            returned_name = get_active_template('example.html')

        self.assertEqual(overriden, template)
        self.assertEqual(overriden_name, template)
        self.assertEqual(overriden_nested, nested_template)
        self.assertEqual(overriden_nested_name, nested_template)
        self.assertEqual(returned, template)
        self.assertEqual(returned_name, template)
        # Check template is deactivated when override exited.
        self.assertRaises(TemplateDoesNotExist, get_active_template, 'example.html')

    def test_override_parent(self):
        # Test override correctly changes also parent template
        parent = TemplateModel(name='parent.html')
        template = TemplateModel(name='example.html', parent=parent)
        with override_template(template):
            overriden = get_active_template('example.html')
            overriden_parent = get_active_template('parent.html')

        self.assertEqual(overriden, template)
        self.assertEqual(overriden_parent, parent)
        # Check templates are deactivated when override exited.
        self.assertRaises(TemplateDoesNotExist, get_active_template, 'example.html')
        self.assertRaises(TemplateDoesNotExist, get_active_template, 'parent.html')


class TestTemplateLoader(TestCase):
    engine = Engine()

    def setUp(self):
        patcher = patch('django_secretary.utils.templates._ACTIVE_TEMPLATES', local())
        self.addCleanup(patcher.stop)
        self._active_templates = patcher.start()

        # Use temporary file storage
        tmp_dir = TempDirectory()
        self.addCleanup(tmp_dir.cleanup)
        storage_patcher = self.settings(MEDIA_ROOT=tmp_dir.path)
        self.addCleanup(storage_patcher.disable)
        storage_patcher.enable()
        self.template = create_template(name='example.html', is_active=True, content='Something completely different')

    def test_get_template_sources(self):
        loader = TemplateLoader(self.engine)
        origin = Origin(name=sentinel.name, loader=loader)
        self.assertCountEqual(loader.get_template_sources(sentinel.name), (origin, ))

    def test_get_contents(self):
        loader = TemplateLoader(self.engine)
        origin = Origin(name='example.html')
        self.assertEqual(loader.get_contents(origin), 'Something completely different')

    def test_get_contents_active(self):
        template = create_template(name='active.html', content='Something else')
        loader = TemplateLoader(self.engine)
        origin = Origin(name='active.html')
        with override_template(template):
            self.assertEqual(loader.get_contents(origin), 'Something else')

    def test_get_contents_not_found(self):
        loader = TemplateLoader(self.engine)
        origin = Origin(name='unknown.html')
        self.assertRaises(TemplateDoesNotExist, loader.get_contents, origin)


class TemplateTest(RenderTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        params = {'NAME': 'django', 'DIRS': [], 'APP_DIRS': False,
                  'OPTIONS': {'loaders': ['django_secretary.utils.templates.TemplateLoader']}}
        self.backend = DjangoTemplates(params)

    def test_template(self):
        model = create_template(name='example.txt', content='Gazpacho!')
        template = Template(model, self.backend)

        self.assertEqual(template.template.name, 'example.txt')
        self.assertEqual(template.template.engine, self.backend.engine)

    def test_template_active(self):
        # Test template returns active template.
        model = create_template(name='example.txt', content='Gazpacho!')
        # Create another model with the same name.
        create_template(name='example.txt', is_active=False, content='Other!')
        template = Template(model, self.backend)

        self.assertEqual(template.template.name, 'example.txt')
        self.assertEqual(template.template.source, 'Gazpacho!')

    def test_template_cached(self):
        # Test template property is cached.
        model = create_template(name='example.txt', content='Gazpacho!')
        template = Template(model, self.backend)

        self.assertIs(template.template, template.template)

    def test_origin(self):
        model = create_template(name='example.txt', content='Gazpacho!')
        template = Template(model, self.backend)

        self.assertEqual(template.origin.name, 'example.txt')

    def test_render(self):
        # Test rendering simple template.
        model = create_template(name='example.txt', content="Hello, I'm {{ name }}.")
        template = Template(model, self.backend)
        context = {'name': 'Kryten'}

        self.assertEqual(template.render(context), "Hello, I'm Kryten.")

        self.assertEqual(self.template_rendered['templates'][0], template.template)
        self.assertEqual(self.template_rendered['context'][0]['name'], 'Kryten')
        self.assertEqual(self.template_rendered['context'][0]['template'], model)

    def test_render_parent(self):
        # Test rendering the template with parent template.
        parent = create_template(name='base.txt', content="Hello, {% block content %}{% endblock %}")
        model = create_template(
            name='example.txt', parent=parent,
            content="{% extends template.parent.name %}{% block content %}I'm {{ name }}.{% endblock %}")
        template = Template(model, self.backend)
        context = {'name': 'Kryten'}

        self.assertEqual(template.render(context), "Hello, I'm Kryten.")

        self.assertEqual(self.template_rendered['templates'][0], template.template)
        self.assertEqual(self.template_rendered['context'][0]['name'], 'Kryten')
        self.assertEqual(self.template_rendered['context'][0]['template'], model)
        self.assertEqual(self.template_rendered['templates'][1].name, 'base.txt')
        self.assertEqual(self.template_rendered['context'][1]['name'], 'Kryten')

    def test_render_invalid(self):
        # Test rendering invalid template.
        model = create_template(name='example.txt', content="Hello, I'm {% invalid %}")
        template = Template(model, self.backend)

        with self.assertRaisesMessage(TemplateSyntaxError, "Invalid block tag on line 1: 'invalid'."):
            template.render({})
